import numpy as np
import scipy as sp
import scipy.stats
import matplotlib.patches as mpatches

def pca_uncertainty_plot(ax,
                         series_colors,
                         classes,
                         scores,
                         bootstrap_classes,
                         bootstrap_scores,
                         class_names=[],
                         x_comp=0,y_comp=1,
                         boot_scatter=False,
                         sample_ellipses=False,
                         boot_group_ellipses=False,
                         group_ellipses=False,
                         sample_bars=False,
                         legend=False,
                         group_tags=False,
                        ):
    #Create the figure
    #fig,axes = plt.subplots(1,2,figsize=(12,6))
    ellipses = []
    F_alpha = 0.05 # 95% confidence limit
    
    for class_num,color in enumerate(series_colors):
        #Get the column that defines the samples as being in this class
        sample_class = classes[:,class_num]
        #Convert to a mask
        class_mask = sample_class == 1        
        #PCA scores
        x_pl = scores[class_mask,x_comp]
        y_pl = scores[class_mask,y_comp]
        ax.scatter(x_pl,y_pl,c=color,edgecolors='k',s=30,zorder=5)
        
        #Number of samples in this class
        num_samples = np.count_nonzero(class_mask)
        
        if bootstrap_classes is not None:
            #Get the column that defines the samples as being in this class
            sample_class = bootstrap_classes[:,class_num]
            #Change to a mask
            class_mask_boot = sample_class == 1
            
            #Bootstrap scores
            x_sc = bootstrap_scores[class_mask_boot,x_comp]
            y_sc = bootstrap_scores[class_mask_boot,y_comp]
        
        #Bootstrap scores scatter plot
        if boot_scatter: ax.scatter(x_sc,y_sc,c=color,edgecolors='none',s=1,zorder=1)
        
        #Error bars on individual samples from bootstrap 
        if sample_bars or sample_ellipses:
            for point in range(num_samples):
                x_pt = x_sc[point::num_samples]
                y_pt = y_sc[point::num_samples]
                if sample_bars:
                    x_err = (np.percentile(x_pt,97.5) - np.percentile(x_pt,2.5))/2
                    y_err = (np.percentile(y_pt,97.5) - np.percentile(y_pt,2.5))/2

                    ax.errorbar(x_pl[point],y_pl[point],xerr=x_err,yerr=y_err,zorder=3,color='k')
                if sample_ellipses:
                    #Calculate the covariance matrix for the bootstrap scores for this sample
                    cov = np.cov(x_pt,y_pt)
                    #Calculate the eigenvalues and eigenvectors of cov
                    lam,v = np.linalg.eig(cov)
                    #Calculate the size and orientation of the confidence ellipse
                    lam = np.sqrt(lam)
                    theta = np.degrees(np.arctan2(*v[:,0][::-1]))
                    
                    pca_samples = len(x_pt)
                    
                    #Calculate the 95% confidence limit based on the F distribution
                    df1 = 2
                    df2 = pca_samples - df1
                    F_val = sp.stats.f.isf(F_alpha,df1,df2)

                    F_mult = F_val * df1*(pca_samples-1)/(df2)
                    F_mult = np.sqrt(F_mult)
                    ellipse_dict = dict(xy=(np.mean(x_pt),np.mean(y_pt)),
                            width=lam[0]*2*F_mult,height=lam[1]*2*F_mult,
                            angle=theta,linewidth=1,linestyle='--',facecolor='none',edgecolor=color,zorder=4)
                    ell = mpatches.Ellipse(**ellipse_dict)
                    if boot_scatter: ell.set_edgecolor('0.2')
                    if boot_group_ellipses: ell.set_edgecolor('0.6')
                    
                    ax.add_artist(ell)
        
        if boot_group_ellipses and np.count_nonzero(class_mask_boot):
            cov = np.cov(x_sc,y_sc)
            lam,v = np.linalg.eig(cov)
            lam = np.sqrt(lam)
            theta = np.degrees(np.arctan2(*v[:,0][::-1]))
            
            pca_samples = len(x_sc)
            
            df1 = 2
            df2 = pca_samples - df1
            F_val = sp.stats.f.isf(F_alpha,df1,df2)

            F_mult = F_val * df1*(pca_samples-1)/(df2)
            F_mult = np.sqrt(F_mult)
            ellipse_dict = dict(xy=(np.mean(x_sc),np.mean(y_sc)),
                                width=lam[0]*2*F_mult,height=lam[1]*2*F_mult,
                                angle=theta,linewidth=2,linestyle='--',facecolor='none',edgecolor=color,zorder=4)
            ell = mpatches.Ellipse(**ellipse_dict)
            if boot_scatter: ell.set_edgecolor('k')
            ax.add_artist(ell)
        
        if group_tags:
            if not boot_group_ellipses:
                cov = np.cov(x_sc,y_sc)
                lam,v = np.linalg.eig(cov)
                lam = np.sqrt(lam)
                theta = np.degrees(np.arctan2(*v[:,0][::-1]))
                pca_samples = len(x_sc)
                df1 = 2
                df2 = pca_samples - df1
                F_val = sp.stats.f.isf(F_alpha,df1,df2)                
                F_mult = F_val * df1*(pca_samples-1)/(df2)
                F_mult = np.sqrt(F_mult)
            x_text_offset = lam[0]*F_mult*-1.1
            y_text_offset = lam[1]*F_mult*-1.1
            ax.text(np.mean(x_sc+x_text_offset),np.mean(y_sc+y_text_offset),
                       class_names[class_num],ha='center',va='center',
                       bbox=dict(edgecolor=color,facecolor='white', alpha=0.9))
            
        if group_ellipses: 
            #Calculate the covariance matrix for the bootstrap scores for this sample
            cov = np.cov(x_pl,y_pl)
            #Calculate the eigenvalues and eigenvectors of cov
            lam,v = np.linalg.eig(cov)
            
            #Calculate the size and orientation of the confidence ellipse
            lam = np.sqrt(lam)
            theta = np.degrees(np.arctan2(*v[:,0][::-1]))
            
            df1 = 2
            df2 = num_samples - df1

            F_val = sp.stats.f.isf(F_alpha,df1,df2)

            F_mult = F_val * df1*(num_samples-1)/(df2)
            F_mult = np.sqrt(F_mult)
            ellipse_dict = dict(xy=(np.mean(x_pl),np.mean(y_pl)),
                                width=lam[0]*2*F_mult,height=lam[1]*2*F_mult,
                                angle=theta,linewidth=2,zorder=2,facecolor='none',edgecolor=color)
            ell = mpatches.Ellipse(**ellipse_dict)
            if boot_scatter: ell.set_edgecolor('k')
            ax.add_artist(ell)
    
    ax.tick_params(labelsize=12)
    if legend:
        patches = []
        for class_name,color in zip(class_names,series_colors):
            patches += [mpatches.Patch(facecolor=color,edgecolor='k',label=class_name)]
        ax.legend(handles=patches)
def score_classes(plsax,
              series_colors,
              classes,
              scores,
              bootstrap_classes,
              bootstrap_scores,
              class_names=[],
              x_comp=0,y_comp=1,
              boot_scatter=False,
              sample_ellipses=False,
              boot_group_ellipses=False,
              group_ellipses=False,
              sample_bars=False,
              legend=False,
              f_alphas=None):
    
    if f_alphas is None:
        f_alphas = np.linspace(0.05,0.95,5)
    #print(f_alphas)
    
    F_scores = []
    
    for class_num,color in enumerate(series_colors):
        #Get the column that defines the samples as being in this class
        sample_class = classes[:,class_num]
        #Convert to a mask
        class_mask = sample_class == 1        
        #PCA scores
        x_pl = scores[class_mask,x_comp]
        y_pl = scores[class_mask,y_comp]
        #plsax.scatter(x_pl,y_pl,c=color,edgecolors='k',s=30,zorder=5)
        
        #Number of samples in this class
        num_samples = np.count_nonzero(class_mask)
        
        #Get the column that defines the samples as being in this class
        sample_class = bootstrap_classes[:,class_num]
        #Change to a mask
        class_mask_boot = sample_class == 1
        
        #Bootstrap scores for this class
        x_sc = bootstrap_scores[class_mask_boot,x_comp]
        y_sc = bootstrap_scores[class_mask_boot,y_comp]
        
        #Calculate the covariance matrix of the boot scores so that we can Hoetelling the samples
        cov = np.cov(x_sc,y_sc)
        icov = np.linalg.inv(cov)
        #print(cov)
        pca_samples = len(x_sc)
        lam,v = np.linalg.eig(cov)
        
        #Calculate the size and orientation of the confidence ellipse
        lam = np.sqrt(lam)
        theta = np.degrees(np.arctan2(*v[:,0][::-1]))
            
        df1 = 2
        df2 = pca_samples - df1
        
        Hoetelling_mult = df1*(pca_samples-1)/(df2)
        
        xy_tuple = (np.mean(x_sc),np.mean(y_sc))
        
        T_center = np.array([xy_tuple])
        #print(T_center.shape)
        
        #print(lam)
        
        for alpha in f_alphas:
            #print(alpha)
            F_val = sp.stats.f.isf(alpha,df1,df2)
            F_mult = F_val * Hoetelling_mult
            F_mult = np.sqrt(F_mult)
            #print(F_mult)
            
            ellipse_dict = dict(xy=xy_tuple,
                                width=lam[0]*2*F_mult,height=lam[1]*2*F_mult,
                                angle=theta,linewidth=1,linestyle=':',zorder=2,facecolor='none',edgecolor=color)
            ell = mpatches.Ellipse(**ellipse_dict)
            plsax.add_artist(ell)
        
        #Iterate over the sample points and calculate the Hoetelling distance and the corresponding survival function
        for xp,yp in zip(x_pl,y_pl):
            sample_point = np.array([[xp,yp]])
            #print(sample_point.shape)
            
            t_rel = sample_point - T_center
            dist = np.sqrt(np.dot(t_rel,t_rel.T))
            #print(dist)
            
            t_right = np.dot(icov,t_rel.T)
            #print(t_right)
            
            t2_distance = np.dot(t_rel,t_right)
            t2_F = t2_distance / Hoetelling_mult
            F_score = sp.stats.f.sf(t2_F,df1,df2)
            F_scores += [F_score]
            #print(F_score)
#             if class_num == 1:
#                 print(sample_point,T_center)
#                 print(t_rel,dist,t2_F,np.sqrt(t2_distance),F_score)
#                 print()
    #print(F_scores)
    F_scores = np.concatenate(F_scores,axis=0)
    return F_scores